/*
 *  $Id: start.c,v 1.13 1994/07/18 19:16:12 bridges Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char vcid[] = "$Id: start.c,v 1.13 1994/07/18 19:16:12 bridges Exp $";
#endif /* lint */

#include "mpiimpl.h"
#include "mpisys.h"

/* NOTES 
   We mark all sends and receives as non-blocking because that is safe here;
   unfortunately, we don't have enough information in the current send/recv
   handle to determine if we are blocking or not.
 */

/*@
    MPI_Start - Initiates a communication with a persistent request handle

Input Parameter:
. request - communication request (handle) 

@*/
int MPI_Start( request )
MPI_Request *request;
{
    int is_available;
    int errno = MPI_SUCCESS;
    MPIR_SHANDLE *shandle;
    MPIR_RHANDLE *rhandle;

    if (MPIR_TEST_REQUEST(MPI_COMM_WORLD,*request))
	return MPIR_ERROR(MPI_COMM_WORLD, errno, "Error in MPI_START" );

    /* See also send.c */
    if ((*request)->type == MPIR_SEND)
    {
      MPIR_Send_setup(request);
      /* device will post the send */
      MPID_post_send( &((*request)->shandle) );
    }
    else if ((*request)->type == MPIR_RECV)
    {

      MPIR_Receive_setup(request);
        /* device will handle queueing of MPIR recv handle in posted-recv 
	   queue         
	 */
        rhandle = &(*request)->rhandle;
        MPID_set_recv_is_nonblocking( &(rhandle->dev_rhandle), 1 );
	MPID_post_recv( rhandle, &(is_available) ); 
    }
    else
	errno = MPIR_ERROR(MPI_COMM_WORLD,MPI_ERR_INTERN,
		   "bad request type in MPI_START");
    return errno;
}


